//
// Sample showing how to use bump mapping in Cg programs
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include	"libExt.h"

#include	<glut.h>
#include	<stdio.h>
#include	<stdlib.h>

#include	"libTexture.h"
#include	"Vector3D.h"
#include	"Vector2D.h"
#include	"Vector4D.h"
#include	"CgProgram.h"
#include	"Mesh.h"
#include	"MeshUtils.h"

Vector3D	eye   ( 7, 5, 7 );			// camera position
Vector3D	light ( 5, 0, 4 );			// light position
float	 	angle = 0;
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
unsigned	diffuseMap;
unsigned	bumpMap;
Mesh      * torus;
CgProgram	vertexProgram   ( CgProgram :: CgVertexProgram   );
CgProgram	fragmentProgram ( CgProgram :: CgFragmentProgram );

void init ()
{
	glClearColor ( 0.0, 0.0, 0.0, 1.0 );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LEQUAL );

	glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
	glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void display ()
{
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

												// draw the light
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ();

	glTranslatef       ( light.x, light.y, light.z );
	glColor4f          ( 1, 1, 1, 1 );
	glutSolidSphere    ( 0.1f, 15, 15 );
	glPopMatrix        ();

	glMatrixMode   ( GL_MODELVIEW );
	glPushMatrix   ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );
	
	vertexProgram.enable   ();
	vertexProgram.bind     ();
	fragmentProgram.enable ();
	fragmentProgram.bind   ();
	
	vertexProgram.setMatrixParameter ( "mv",  CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_IDENTITY );
	vertexProgram.setMatrixParameter ( "mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY );
	vertexProgram.setMatrixParameter ( "mvi", CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_INVERSE_TRANSPOSE   );

	torus -> render ();
	
	vertexProgram  .disable ();
	fragmentProgram.disable ();

    glPopMatrix ();

	glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   											// factor all camera ops into projection matrix
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					0, 0, 0,				// center
					0, 0, 1 );				// up

   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
}

void	animate ()
{
	angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );

	light.x = 2*cos ( angle );
	light.y = 2*sin ( angle );
	light.z = 3 + 0.3 * sin ( angle / 3 );

	vertexProgram.bind         ();
	vertexProgram.setParameter ( "eyePos",   Vector4D ( eye,   1 ) );
	vertexProgram.setParameter ( "lightPos", Vector4D ( light, 1 ) );
	vertexProgram.unbind       ();

	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( 500, 500 );

								// create window
	glutCreateWindow ( "Bumpmapping with Cg" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

	init           ();
	initExtensions ();

	if ( !vertexProgram.load ( "bump-vertex.cg" ) )
	{
		printf ( "Error compling vertex program:%s\n", vertexProgram.getErrorDescription ().c_str () );

		exit ( 1 );
	}

	if ( !fragmentProgram.load ( "bump-fragment.cg" ) )
	{
		printf ( "Error compling fragment program:%s\n", fragmentProgram.getErrorDescription ().c_str () );

		exit ( 1 );
	}

	diffuseMap  = createTexture2D ( true, "../../Textures/oak.bmp" );
	bumpMap     = createTexture2D ( false, "../../Textures/Bumpmaps/normal2.bmp" );

    									// install program object as part of
    									// current state
	fragmentProgram.enable     ();
	fragmentProgram.bind       ();
    fragmentProgram.setTexture ( "bumpMap",    bumpMap   );
    fragmentProgram.setTexture ( "diffuseMap", diffuseMap );
	fragmentProgram.disable    ();

	torus = makeTorus ( 1, 3, 30, 30 );

	torus -> addCoordAssignment ( Mesh :: tagVertex,   Mesh :: tagVertex );
	torus -> addCoordAssignment ( Mesh :: tagNormal,   Mesh :: tagNormal );
	torus -> addCoordAssignment ( Mesh :: tagTexCoord, Mesh :: tagTex0   );
	torus -> addCoordAssignment ( Mesh :: tagTangent,  Mesh :: tagTex1   );
	torus -> addCoordAssignment ( Mesh :: tagBinormal, Mesh :: tagTex2   );
	torus -> createBuffers      ();

    glutMainLoop ();

	return 0;
}
