//
// Example of glowing object with glow mask via Cg shaders
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#include    <glut.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "TypeDefs.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "boxes.h"
#include    "FrameBuffer.h"
#include    "CgProgram.h"

Vector3D    eye   ( -0.5, -0.5, 1.5 );  // camera position
unsigned    decalMap;                   // decal (diffuse) texture
unsigned    stoneMap;
unsigned    teapotMap;
unsigned	glowMap;
unsigned    screenMap1;
unsigned    screenMap2;
unsigned    screenMap3;
float       angle     = 0;
float       rot       = 0;
bool        useFilter = true;
bool		rotate    = true;

FrameBuffer	buffer1       ( 256, 256, FrameBuffer :: depth32 );
FrameBuffer	buffer2       ( 256, 256 );
FrameBuffer	buffer3       ( 256, 256 );
CgProgram	vertexBlurX   ( CgProgram :: CgVertexProgram   );
CgProgram	fragmentBlurX ( CgProgram :: CgFragmentProgram );
CgProgram	vertexBlurY   ( CgProgram :: CgVertexProgram   );
CgProgram	fragmentBlurY ( CgProgram :: CgFragmentProgram );

void    renderToBuffer ();
void	reshape        ( int w, int h );

void init ()
{
    glClearColor ( 0.0, 0.0, 0.0, 1.0 );
    glEnable     ( GL_DEPTH_TEST );
    glEnable     ( GL_TEXTURE_2D );
    glDepthFunc  ( GL_LEQUAL     );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void drawSceneGlow ()
{
    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glRotatef    ( rot, 0, 0, 1 );

	glColor4f       ( 0.7, 0.7, 0.7, 1 );
    glBindTexture   ( GL_TEXTURE_2D, glowMap );
    glTranslatef    ( 0.2, 1, 1.5 );
    glRotatef       ( angle * 45.3, 1, 0, 0 );
    glRotatef       ( angle * 57.2, 0, 1, 0 );
    glutSolidTeapot ( 0.3 );

	glColor4f       ( 1, 1, 1, 1 );
    glPopMatrix     ();
}

void drawScene ()
{
    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glRotatef    ( rot, 0, 0, 1 );

    drawBox  ( Vector3D ( -5, -5, 0 ), Vector3D ( 10, 10, 3 ), stoneMap, false );
    drawBox  ( Vector3D ( 3, 2, 0.5 ), Vector3D ( 1,  2,  2 ), decalMap );

    glBindTexture   ( GL_TEXTURE_2D, teapotMap );
    glTranslatef    ( 0.2, 1, 1.5 );
    glRotatef       ( angle * 45.3, 1, 0, 0 );
    glRotatef       ( angle * 57.2, 0, 1, 0 );
    glutSolidTeapot ( 0.3 );

    glPopMatrix     ();
}

void    startOrtho ( int size )
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, size, 0, size, -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix
}

void	drawQuad ( unsigned tex, int size )
{
    glBindTexture ( GL_TEXTURE_2D, tex );
    glBegin       ( GL_QUADS );

        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( 1,    0 );
        glVertex2f   ( size, 0 );

        glTexCoord2f ( 1,    1 );
        glVertex2f   ( size, size );

        glTexCoord2f ( 0, 1 );
        glVertex2f   ( 0, size );

    glEnd   ();
}

void display ()
{
	if ( !useFilter )
	{
		glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
		
		drawScene ();
		
		glutSwapBuffers ();
	
		return;
	}
	
    renderToBuffer ();

    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    startOrtho ( 256 );

	buffer2.bind       ();
	vertexBlurX.enable ();
	vertexBlurX.bind   ();

	fragmentBlurX.enable ();
	fragmentBlurX.bind   ();

	vertexBlurX.setMatrixParameter ( "mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY );
	
	drawQuad ( screenMap1, 256 );

	vertexBlurX.  disable ();
	fragmentBlurX.disable ();
	buffer2.unbind ();

	glBindTexture ( GL_TEXTURE_2D, 0 );
	buffer3.bind ();

	vertexBlurY.enable ();
	vertexBlurY.bind   ();

	fragmentBlurY.enable ();
	fragmentBlurY.bind   ();

	vertexBlurY.setMatrixParameter ( "mvp", CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY );
	
	drawQuad ( screenMap2, 256 );

	vertexBlurY.  disable ();
	fragmentBlurY.disable ();
	
	buffer3.unbind ();

    endOrtho ();

	reshape ( 512, 512 );
	
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	drawScene ();

	startOrtho ( 512 );

	glEnable    ( GL_BLEND );
	glBlendFunc ( GL_ONE, GL_ONE );
	glDisable   ( GL_DEPTH_TEST );
	
	drawQuad ( screenMap3, 512 );

	glDisable ( GL_BLEND );
	glEnable  ( GL_DEPTH_TEST );
	
	endOrtho ();

    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,    // eye
                    3, 3, 1,                // center
                    0, 0, 1 );              // up
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );

    if ( key == 'f' || key == 'F' )
        useFilter = !useFilter;
	else
	if ( key == ' ' )
		rotate = !rotate;
}

void    specialKey ( int key, int x, int y )
{
    if ( key == GLUT_KEY_RIGHT )
        rot += 5;
    else
    if ( key == GLUT_KEY_LEFT )
        rot -= 5;

    glutPostRedisplay ();
}

void    animate ()
{
	if ( rotate )
		angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );

    glutPostRedisplay ();
}

void    renderToBuffer ()
{
	glBindTexture ( GL_TEXTURE_2D, 0 );

	buffer1.bind ();

    glClearColor ( 0, 0, 0, 1 );
    glClear      ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    reshape ( buffer1.getWidth (), buffer1.getHeight () );

    drawSceneGlow ();

    buffer1.unbind ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    glutInitWindowSize  ( 512, 512 );

                                // create window
    glutCreateWindow ( "OpenGL glow with mask effect" );

                                // register handlers
    glutDisplayFunc  ( display    );
    glutReshapeFunc  ( reshape    );
    glutKeyboardFunc ( key        );
    glutSpecialFunc  ( specialKey );
    glutIdleFunc     ( animate    );

    init           ();
    initExtensions ();

	if ( !vertexBlurX.load ( "x-blur-vertex.cg" ) )
	{
		printf ( "Error compling vertex program:%s\n", vertexBlurX.getErrorDescription ().c_str () );

		exit ( 1 );
	}

	if ( !fragmentBlurX.load ( "x-blur-fragment.cg" ) )
	{
		printf ( "Error compling fragment program:%s\n", fragmentBlurX.getErrorDescription ().c_str () );

		exit ( 1 );
	}

	if ( !vertexBlurY.load ( "y-blur-vertex.cg" ) )
	{
		printf ( "Error compling vertex program:%s\n", vertexBlurY.getErrorDescription ().c_str () );

		exit ( 1 );
	}

	if ( !fragmentBlurY.load ( "y-blur-fragment.cg" ) )
	{
		printf ( "Error compling fragment program:%s\n", fragmentBlurY.getErrorDescription ().c_str () );

		exit ( 1 );
	}

	screenMap1 = buffer1.createColorTexture ();
	screenMap2 = buffer2.createColorTexture ();
	screenMap3 = buffer3.createColorTexture ();

	buffer1.create ();
	buffer1.bind   ();
	buffer1.attachColorTexture ( GL_TEXTURE_2D, screenMap1 );
	buffer1.unbind ();

	buffer2.create ();
	buffer2.bind   ();
	buffer2.attachColorTexture ( GL_TEXTURE_2D, screenMap2 );
	buffer2.unbind ();

	buffer3.create ();
	buffer3.bind   ();
	buffer3.attachColorTexture ( GL_TEXTURE_2D, screenMap3 );
	buffer3.unbind ();

    decalMap  = createTexture2D ( true, "../../Textures/oak.bmp" );
    stoneMap  = createTexture2D ( true, "../../Textures/block.bmp" );
    teapotMap = createTexture2D ( true, "../../Textures/Oxidated.jpg" );
	glowMap   = createTexture2D ( true, "glow-mask.bmp" );

	fragmentBlurX.enable     ();
	fragmentBlurY.bind       ();
	fragmentBlurX.setTexture ( "tex0", screenMap1  );
	fragmentBlurX.disable    ();

	fragmentBlurY.enable     ();
	fragmentBlurY.bind       ();
	fragmentBlurY.setTexture ( "tex0", screenMap2  );
	fragmentBlurY.disable    ();

    printf ( "Press F key to turn filtering on/off\nSpace toggles animation" );

    glutMainLoop ();

    return 0;
}
